/*---------------------------------------------------------------------.
| zflate.c							       |
| 								       |
|  Copyright (c) 1997 Jonathan Hudson				       |
| 								       |
|  You may distribute under the terms as specified in zflate.txt       |
`---------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <zlib.h>
#include <string.h>

#ifdef QDOS
#include <qdos.h>

#define VERSION " 0.01"
# ifdef DEFLATE
  char _prog_name[] = "deflate" VERSION;
# else
  char _prog_name[] = "inflate" VERSION;
# endif
# define OPEN io_open
# define CLOSE io_close
# define WRITE(f,b,n) io_sstrg(f, -1, b, n)
# define READONLY 1
# define WRITEONLY 3
# define ENOMEM ERR_OM
# define ERROR ERR_BP
# define FERROR -16
int READ(int f, void *b, int n)
{
    int res = io_fstrg(f, -1, b, n);
    return (res == ERR_EF) ? 0 : res;
}
#else
/* Use a sensible operating system */
# define OPEN open
# define CLOSE close
# define READ read
# define WRITE write
# define READONLY O_RDONLY
# define WRITEONLY O_CREAT|O_WRONLY|O_TRUNC,0666
# define DOINFLATE
# define DODEFLATE
# define ERROR 255
# define FERROR errno 
#endif

#define MSIZ (16*1024)
#define DEF_MEM_LEVEL 8 

typedef struct
{
    int file;
    void *buf;
    int err;
    z_stream stream;
    void *ptr;
    unsigned long crc;
} ZFILE;


static void *xalloc (size_t n)
{
    void *p;
    if ((p = malloc (n)) == NULL)
    {
	exit (ENOMEM);
    }
    return p;
}
#ifdef QDOS
static int grabparams (char *sp, short nc, char sep, char **ib, int *f1)
{
    int i, j = 0, k;
    char *p;
    *ib = NULL;
    *f1 = 0;

    for (p = sp, i = 0; i < nc; i++, p++)
    {
	if (*p == sep)
	{
	    if (j == 0)
	    {
		*p = 0;
		if ((k = strtoul (sp, NULL, 0)) == 0)
		{
		    *ib = sp;
		    i++;
		    break;
		}
		else
		{
		    *ib = (char *) k;
		    sp = p + 1;
		    j = 1;
		}
	    }
	    else
	    {
		i++;
		break;
	    }
	}
    }
    if (j)
    {
	*p = 0;
	*f1 = strtoul (sp, NULL, 0);
    }
    return i;
}
#endif

#ifdef INFLATE
static int zread (ZFILE * s, void *buf, int len)
{
    if (s->err == Z_DATA_ERROR || s->err == Z_ERRNO)
	return -1;
    if (s->err == Z_STREAM_END)
	return 0;		/* EOF */

    s->stream.next_out = buf;
    s->stream.avail_out = len;

    while (s->stream.avail_out != 0)
    {
	if (s->stream.avail_in == 0)
	{
	    s->stream.avail_in = READ(s->file,s->buf,MSIZ);
	    s->stream.next_in = s->buf;
	}
	s->err = inflate (&(s->stream), Z_NO_FLUSH);
	if (s->err != Z_OK)
	    break;
    }
    return (int) (len - s->stream.avail_out);
}
#endif

#ifdef DEFLATE
static int zwrite (ZFILE * s, void *buf, int len)
{
    s->stream.next_in = buf;
    s->stream.avail_in = len;

    while (s->stream.avail_in != 0)
    {
	if (s->stream.avail_out == 0)
	{
	    s->stream.next_out = s->buf;
	    if (WRITE(s->file,s->buf,MSIZ) != MSIZ)
	    {
		s->err = Z_ERRNO;
		break;
	    }
	    s->stream.avail_out = MSIZ;
	}
	s->err = deflate (&(s->stream), Z_NO_FLUSH);
	if (s->err != Z_OK)
	    break;
    }
    return (int) (len - s->stream.avail_in);
}

int zflush (ZFILE * s, int flush)
{
    int len;
    int done = 0;

    s->stream.avail_in = 0;	/* should be zero already anyway */

    for (;;)
    {

	len = MSIZ - s->stream.avail_out;
	if (len != 0)
	{
	    if (WRITE(s->file,s->buf,len) != len)
	    {
		s->err = Z_ERRNO;
		return Z_ERRNO;
	    }
	    s->stream.next_out = s->buf;
	    s->stream.avail_out = MSIZ;
	}
	if (done)
	    break;

	s->err = deflate (&(s->stream), flush);
	done = (s->stream.avail_out != 0 || s->err == Z_STREAM_END);
	if (s->err != Z_OK && s->err != Z_STREAM_END)
	    break;
    }
    return s->err == Z_STREAM_END ? Z_OK : s->err;
}
#endif

static void putlsbL (unsigned char z[4], unsigned long x)
{
    short n;
    for (n = 0; n < 4; n++) {
        z[n] = (x & 0xff);
        x >>= 8;
    }
}

#define USE_MEM 1
#define USE_FILE 0

#if 0 && defined(QDOS)
#include <stdarg.h>
#define C1 0x00010001
void report (int chid, char *ctl,...)
{
    va_list va;
    char t[80];

    va_start (va, ctl);
    vsprintf (t, ctl, va);
    va_end (va);
    io_sstrg (chid, -1, t, (short) strlen (t));
}   
#endif

int main (int ac, char **av)
{
    int f = 0;
    ZFILE *s;
    char *buf, *inbuf, *outbuf;
    char *fn1,*fn2;
    int n, buflen, outlen, res = 0;
    short intype = 0;
    short outtype = 0;
    short done;
    short usegz = 0;
#ifdef QDOS
    int oc = -1;
    int f1,f2,ch1,ch2;

    s = xalloc (sizeof (ZFILE));
    memset (s, 0, sizeof (ZFILE));
    {
	short m, nc, nchan;
	char *sp, *fnm, sep;
	long *sptr;

	ch1 = ch2 = 0;

	sp = *(av + 1);		 /* The QDOS stack */
	nchan = *((short *) sp); /* num chans      */
	sp += 2;
	
	if (nchan)
	{
	    sptr = (long *)sp;
	}
	sp = sp + (nchan << 2);
	
	nc = *((short *) sp);
	sp += 2;
	
	if (nc)
	{
	    char *p = sp;
	    short i;
	    
	    sep = *p;
	    
	    for (i = 0; i < nc - 1; i++, p++)
	    {
		*p = *(p + 1);
	    }
	    *p = sep;

	    if(*sp == '-' && nchan >= 2)
	    { 
		if(*(sp+1) & 1)
		{
		    ch1 = *(sptr+0);
		}
		if(*(sp+1) & 2)
		{
		    ch2 = *(sptr+1);
		}
		nchan -= 2;
		sptr += 2;
		if(*(sp+1) & 4)
		{
		    usegz = 1;
		    s->crc = crc32(0L, Z_NULL, 0);
		}
		sp += 3;
	    }
	    if(nchan == 1)
	    {
		oc = *sptr;
	    }

	    if(ch1 == 0)
	    {
		m = grabparams (sp, nc, sep, (char **) &inbuf, &buflen);
		if(buflen) 
		{
		    intype = USE_MEM;
		}
		else
		{
		    fn1 = inbuf;
		}
	    }
	    else m = 0;

	    if(ch2 == 0)
	    {
		grabparams (sp + m, nc - m, sep, (char **) &outbuf, &outlen);
		if(outlen) 
		{
		    outtype = USE_MEM;
		}
		else
		{
		    fn2 = outbuf;
		}
	    }
	}
    }
#else
    fn1 = *(av+1);
    fn2 = *(av+2);
#endif

#ifndef QDOS
    if (strcmp ("deflate", av[0]) == 0)
#endif

#ifdef DEFLATE
    {
	if (intype == USE_FILE)
	{
#ifdef QDOS
	    if(ch1) 
		f = ch1;
	    else
#endif
		f = OPEN(fn1,READONLY);
	    inbuf = xalloc (MSIZ);
	}
	else
	{
	    inbuf;
	    n = buflen;
	}

	if (outtype == USE_FILE)
	{
#ifdef QDOS
	    if(ch2)
		s->file = ch2;
	    else
#endif
		s->file = OPEN(fn2,WRITEONLY);
	    s->buf = xalloc (MSIZ);
	    s->stream.avail_out = MSIZ;
	}
	else
	{
	    s->buf = outbuf;
	    s->stream.avail_out = outlen;
	}

	if (f < 0 || s->file < 0)
	{
	    res = ERROR;
	}
	else
	{
	    if(usegz)
	    {
		deflateInit2(&(s->stream), Z_DEFAULT_COMPRESSION,
                           Z_DEFLATED, -MAX_WBITS, DEF_MEM_LEVEL, 0);
	    }
	    else
		deflateInit (&s->stream, Z_DEFAULT_COMPRESSION);
	    s->stream.next_out = s->buf;

	    for (done = 0; !done;)
	    {
		if (intype == USE_FILE)
		{
		    n = READ(f,inbuf,MSIZ);
		    if (n <= 0)
		    {
			done = 1;
			continue;
		    }
		}
		else
		{
		    done = 1;
		}
		
		if (outtype == USE_FILE)
		{
		    zwrite (s, inbuf, n);
		}
		else
		{
		    s->stream.next_in = (unsigned char *)inbuf;
		    s->stream.avail_in = n;
		    s->err = deflate (&(s->stream), Z_NO_FLUSH);
		}
		if(usegz) s->crc = crc32(s->crc, (void *)inbuf, n);
		if(s->err < 0 || n < 0)
		{
		    res = ERROR;
		    break;
		}
	    }
	    
	    if (intype == USE_FILE)
	    {
#ifdef QDOS
		if(ch1 == 0)
#endif
		    CLOSE(f);
	    }
	    if (outtype == USE_FILE)
	    {
		zflush (s, Z_FINISH);
		if(usegz)
		{
		    unsigned char x[8];
		    putlsbL (x, s->crc);
		    putlsbL (x+4, s->stream.total_in);
		    WRITE(s->file, x, 8);
		}
#ifdef QDOS
		if(ch2 == 0)
#endif
		CLOSE(s->file);
	    }
	    else
	    {
		s->err = deflate (&(s->stream), Z_FINISH);
		if(usegz)
		{
		    putlsbL(s->stream.next_out,  s->crc);
		    putlsbL(s->stream.next_out+4, s->stream.total_in);
		}
	    }
	}
    }
#endif

#ifndef QDOS
    if (strcmp ("inflate", av[0]) == 0)
#endif

#ifdef INFLATE
    {
	if (intype == USE_FILE)
	{
#ifdef QDOS
	    if(ch1) 
		s->file = ch1;
	    else
#endif
		s->file = OPEN(fn1,READONLY);
	    s->stream.next_in = s->buf = xalloc (MSIZ);
	}
	else
	{
	    s->stream.next_in = (unsigned char *) inbuf;
	    s->stream.avail_in = buflen;
	}

	if (outtype == USE_FILE)
	{
#ifdef QDOS
	    if(ch2) 
		f = ch2;
	    else
#endif
	    f = OPEN(fn2,WRITEONLY);
	    s->stream.avail_out = MSIZ;
	    buf = outbuf = xalloc (MSIZ);
	}
	else
	{
	    s->stream.avail_out = n = outlen;
	    s->stream.next_out = (unsigned char *)outbuf;
	    buf = outbuf;
	}
	if (f < 0 || s->file < 0)
	{
	    res = FERROR;
	}
	else
	{
	    if(usegz)
		inflateInit2(&(s->stream), -MAX_WBITS);
	    else
		inflateInit (&(s->stream));

	    for (done = 0; done == 0;)
	    {
		if (intype == USE_FILE)
		{
		    if ((n = zread (s, buf, MSIZ)) <= 0)
		    {
			done = 1;
		    }
		}
		else
		{
		    if (outtype == USE_FILE)
		    {
			s->stream.next_out = (unsigned char *) buf;
			s->stream.avail_out = n = MSIZ;
		    }

		    {
			s->err = inflate (&s->stream, Z_NO_FLUSH);
			done = (s->err == Z_STREAM_END);
		    }
		    n -= s->stream.avail_out;
		}
		if(s->err < 0 || n < 0)
		{
		    res = ERROR;
		    break;
		}
		if (outtype == USE_FILE)
		{
		    WRITE(f,buf,n);
		}
		else
		{
		    buf = buf + n;
		}
	    }
	    if (intype == USE_FILE)
	    {
#ifdef QDOS
		if(ch1 == 0)
#endif
		    CLOSE(s->file);
	    }
	    if (outtype == USE_FILE)
	    {
#ifdef QDOS
		if(ch2== 0)
#endif
		CLOSE(f);
	    }
	}
    }
#endif

#ifdef QDOS
    if (oc != -1)
    {
	char rtext[20];
	int n;
	if (res == 0)
	    res = s->stream.total_out;
	itoa (res, rtext);
	n = strlen (rtext);
	*(rtext + n) = '\n';
	io_sstrg (oc, -1, rtext, n + 1);
    }
#endif
    return res;
}
#ifdef QDOS
int (*_Cstart) (int, char **) = main;
#endif



